#Modified from Enzo Busseti 2017-2019.
import unittest
import numpy as np
import coneref

HOW_MANY_DERIVATIVE_SAMPLES = 1000
HOW_LONG_DERIVATIVE_TEST_STEP = 1e-6


def size_vec(x):
    return 1 if isinstance(x, float) else len(x)


class BaseTestCone(unittest.TestCase):

    """Base class for cones tests."""

    sample_vecs = []
    sample_vec_proj = []
    sample_vecs_are_in = []
    sample_vecs_are_diff = []

    def make_cache(self, n):
        if self.test_cone == "SOC":
            return np.zeros(n)
        elif self.test_cone == "EXP":
            return np.zeros((n))
        elif self.test_cone == "EXP_DUAL":
            return np.zeros((n))
        elif self.test_cone == "PSD":
            eval_cache = np.zeros(np.sum(n))
            evec_cache = np.zeros(np.sum(n**2))
            return [eval_cache, evec_cache]

    def Pi(self, x, cache):
        if self.test_cone == "SOC":
            return coneref.SOC_Pi(x, cache)
        elif self.test_cone == "PSD":
            return coneref.PSD_Pi(x, cache[0], cache[1])
        elif self.test_cone == "EXP":
            return coneref.exp_primal_Pi(x, cache)
        elif self.test_cone == "EXP_DUAL":
            return coneref.exp_dual_Pi(x, cache)
   
    def diff(self, x, delta, cache):
        #print("DIFF FUNCTION")
        if self.test_cone == "SOC":
            diff_operator = coneref.SOC_Pi_diff(x, cache)
            #print("Differentiating SOC")
        elif self.test_cone == "PSD":
            diff_operator = coneref.SDP_Pi_diff(x, cache[0], cache[1])
        elif self.test_cone == "EXP":
            diff_operator = coneref.exp_primal_Pi_diff(x, cache)
        elif self.test_cone == "EXP_DUAL":
            diff_operator = coneref.exp_dual_Pi_diff(x, cache)    
        return diff_operator.matvec(delta)

    def test_contains(self):
        for x, isin in zip(self.sample_vecs, self.sample_vecs_are_in):
             cache = self.make_cache(len(x))
             res = self.Pi(x, cache)
             Pix = res
             self.assertTrue(np.alltrue(Pix == x) == isin)

    def test_proj(self):
        for x, proj_x in zip(self.sample_vecs, self.sample_vec_proj):
             cache = self.make_cache(len(x))
             Pix = self.Pi(x, cache)
             if not np.allclose(Pix, proj_x):
                print("Pix:", Pix)
                print("proj_x:", proj_x)
                print("x:", x)
                print(" ")
             self.assertTrue(np.allclose(Pix, proj_x))
        


    def test_derivative_random(self):
        for x, isdiff in zip(self.sample_vecs,
                             self.sample_vecs_are_diff):
            x = np.random.randn(len(x))
            cache = self.make_cache(len(x))
            proj_x = self.Pi(x, cache)

            # Due to Pybind11-properties we must (for the test) fill the cache manually.
            if self.test_cone == "SOC":
                cache = coneref.SOC_Pi(x, cache)
            elif self.test_cone == "PSD":
                X = coneref.matrix_from_lower_triangular(x)
                evals, evecs = np.linalg.eigh(X)
                cache[0] = evals
                cache[1] = evecs.flatten('F')
            elif self.test_cone == "EXP":
                proj = coneref.exp_primal_Pi(x, cache)
                cache[0] = proj[0]
                cache[1] = proj[1]
                cache[2] = proj[2] - x[2]
            elif self.test_cone == "EXP_DUAL":
                # Filling this cache manually is complicated. We use a hack.
                dict_of_cones = {'ed': 1}
                q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
                  = coneref.make_prod_cone_cache(dict_of_cones)
                cones = coneref.parse_cone_dict(dict_of_cones)
                _cones_parsed = coneref.parse_cone_dict_cpp(cones)
                return_info = coneref.prod_cone_Pi_return_cache(x, _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
                cache = return_info[5]
            
            for i in range(HOW_MANY_DERIVATIVE_SAMPLES):
                delta = np.random.randn(
                    size_vec(x)) * HOW_LONG_DERIVATIVE_TEST_STEP*0.1
                # new_cache is not used later.
                new_cache = self.make_cache(len(x))
                proj_x_plus_delta = self.Pi(x + delta, new_cache)
 
                dproj_x = self.diff(x, delta, cache)
                if not np.allclose(proj_x + dproj_x, proj_x_plus_delta, atol = 1e-5):
                    print('x:', x)
                    print('Pi x:', proj_x)
                    print('delta:')
                    print(delta)
                    print('Pi (x + delta) - Pi(x):')
                    print(proj_x_plus_delta - proj_x)
                    print('DPi delta:')
                    print(dproj_x)
                    print("proj_x + dproj_x:")
                    print(proj_x + dproj_x)
                    print("proj_x_plus_delta:")
                    print(proj_x_plus_delta)

                self.assertTrue(np.allclose(
                    proj_x + dproj_x,
                    proj_x_plus_delta, atol = 1e-5))


class TestExpPri(BaseTestCone):

    test_cone = "EXP"
    sample_vecs = [np.array([0., 0., 0.]),
                   np.array([-10., -10., -10.]),
                   np.array([10., 10., 10.]),
                   np.array([1., 2., 3.]),
                   np.array([100., 2., 300.]),
                   np.array([-1., -2., -3.]),
                   np.array([-10., -10., 10.]),
                   np.array([1., -1.,  1.])]
                   #np.array([0.08755124, -1.22543552, 0.84436298])]
    sample_vec_proj = [np.array([0., 0., 0.]),
                       np.array([-10., 0., 0.]),
                       np.array([4.26306172,  7.51672777, 13.25366605]),
                       np.array([0.8899428, 1.94041882, 3.06957225]),
                       np.array([73.77502858,  33.51053837, 302.90131756]),
                       np.array([-1., 0., 0.]),
                       np.array([-10., 0., 10.]),
                       np.array([0.22972088, 0.09487128, 1.06839895])]
                       #np.array([3.88378507e-06, 2.58963810e-07, 0.84436298])]  
    sample_vecs_are_in = [True, False, False,
                          False, False, False, False]#, False]
    sample_vecs_are_diff = [False, True, True, True, True, True, True]#, True]


class TestExpDua(BaseTestCone):

    test_cone = "EXP_DUAL"
    sample_vecs = [np.array([0., 0., 0.]),
                   np.array([-1., 1., 100.]),
                   np.array([1., 1., 100.]),
                   np.array([-1., -2., -3.])]
    sample_vec_proj = [np.array([0., 0., 0.]),
                       np.array([-1., 1., 100.]),
                       np.array([0., 1., 100.]),
                       np.array([-0.1100572, -0.05958119,  0.06957226])]
    sample_vecs_are_in = [True, True, False, False]
    sample_vecs_are_diff = [False, True, True, True]


class TestSecondOrder(BaseTestCone):

    test_cone = "SOC"
    sample_vecs = [np.array([1., 0., 0.]),
                   np.array([1., 2., 2.]),
                   np.array([-10., 2., 2.]),
                   np.array([-2 * np.sqrt(2), 2., 2.]),
                   np.array([-1., 2., 2.]),
                   np.array([0., 1.]),
                   np.array([.5, -.5])]
    sample_vec_proj = [np.array([1., 0., 0.]),
                       [(2 * np.sqrt(2) + 1) / (2),
                        (2 * np.sqrt(2) + 1) / (2 * np.sqrt(2)),
                        (2 * np.sqrt(2) + 1) / (2 * np.sqrt(2))],
                       np.array([0, 0, 0]),
                       np.array([0, 0, 0]),
                       np.array([0.9142135623730951,
                                 0.6464466094067263,
                                 0.6464466094067263
                                 ]),
                       np.array([.5, .5]),
                       np.array([.5, -.5])]
    sample_vecs_are_in = [True, False, False, False, False, False, True]
    sample_vecs_are_diff = [True, True, True, False, True, True, False]


class TestProduct(BaseTestCone):

    test_cone = "prod_cone"

    def test_baseProduct(self):

        # Test 1 and 2
        dict_of_cones = {'l': 3}
        q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
         = coneref.make_prod_cone_cache(dict_of_cones)
        _cones = coneref.parse_cone_dict(dict_of_cones)
        _cones_parsed = coneref.parse_cone_dict_cpp(_cones)
        Pix = coneref.prod_cone_Pi(np.arange(3.), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.alltrue(Pix == np.arange(3.)))
        Pix = coneref.prod_cone_Pi(np.array([1., -1., -1.]), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.alltrue(Pix == [1, 0, 0]))

        # Test 3 and 4
        dict_of_cones = {'l': 2, 's': [1]}
        q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
         = coneref.make_prod_cone_cache(dict_of_cones)
        _cones = coneref.parse_cone_dict(dict_of_cones)
        _cones_parsed = coneref.parse_cone_dict_cpp(_cones)
        Pix = coneref.prod_cone_Pi(np.arange(3.), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.alltrue(Pix == range(3)))
        Pix =  coneref.prod_cone_Pi(np.array([1., -1., -1.]), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.alltrue(Pix  == [1, 0, 0]))

        # Test 5
        dict_of_cones = {'s': [2, 1]}
        q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
         = coneref.make_prod_cone_cache(dict_of_cones)
        _cones = coneref.parse_cone_dict(dict_of_cones)
        _cones_parsed = coneref.parse_cone_dict_cpp(_cones)
        Pix = coneref.prod_cone_Pi(np.arange(4.), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.allclose(Pix - np.array(
              [0.20412415, 0.90824829, 2.02062073, 3]), 0.))
        Pix = coneref.prod_cone_Pi(np.array([1, -20., 1, -1]), _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
        self.assertTrue(np.allclose(Pix, np.array([7.57106781, -10.70710678,
                                                    7.57106781,   0.])))

    def test_deriv_Product(self):
        print(" ")
        dict_of_cones = {'z': 2, 'l': 20,  'q': [3, 4],
                    's': [3, 4, 5], 'ep': 10, 'ed': 10}
        m = 22 + 7 + (6 + 10 + 15) + 30 + 30
        q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
         = coneref.make_prod_cone_cache(dict_of_cones)
        _cones = coneref.parse_cone_dict(dict_of_cones)
        _cones_parsed = coneref.parse_cone_dict_cpp(_cones)

        for j in range(100):
            x = np.random.randn(m)
            q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
             = coneref.make_prod_cone_cache(dict_of_cones)

            return_info = coneref.prod_cone_Pi_return_cache(x, _cones_parsed, q_cache, eval_cache, evec_cache, ep_cache, ed_cache)
            proj_x, q_cache, eval_cache, evec_cache, ep_cache, ed_cache = return_info[0], return_info[1], return_info[2], return_info[3], return_info[4], return_info[5]
            diff_operator = coneref.prod_cone_Pi_diff(x, _cones_parsed, q_cache, eval_cache, 
                                 evec_cache, ep_cache, ed_cache)
            for i in range(HOW_MANY_DERIVATIVE_SAMPLES):
                delta = np.random.randn(size_vec(x)) * \
                    HOW_LONG_DERIVATIVE_TEST_STEP
                # print('x + delta:', x + delta)
                q_cache_new, eval_cache_new, evec_cache_new, ep_cache_new, ed_cache_new \
                    = coneref.make_prod_cone_cache(dict_of_cones)
                proj_x_plus_delta = coneref.prod_cone_Pi(x + delta, _cones_parsed, q_cache_new, eval_cache_new, evec_cache_new, ep_cache_new, ed_cache_new)
                dproj_x = diff_operator.matvec(delta)

                if not np.allclose(proj_x + dproj_x, proj_x_plus_delta, atol=1e-5):
                    print(dict_of_cones)
                    print('x:', x)
                    print('Pi (x):', proj_x)
                    # print(delta)
                    print('Pi (x + delta) - Pi(x):')
                    print(proj_x_plus_delta - proj_x)
                    print('DPi delta:')
                    print(dproj_x)

                    print('error:')
                    print(proj_x + dproj_x -
                          proj_x_plus_delta)

                self.assertTrue(np.allclose(
                    proj_x + dproj_x,
                    proj_x_plus_delta, atol=1e-5))


class TestSemiDefinite(BaseTestCone):

    test_cone = "PSD"
    sample_vecs = [np.array([2, 0, 0, 2, 0, 2.]), np.array([1.]), np.array([-1.]),
                   np.array([10, 20., 10.]),
                   np.array([10, 0., -3.]),
                   np.array([10, 20., 0., 10, 0., 10]),
                   np.array([1, 20., 30., 4, 50., 6]),
                   np.array([1, 20., 30., 4, 50., 6, 200., 20., 1., 0.])]
    sample_vec_proj = [[2, 0, 0, 2, 0, 2], [1.], [0.],
                       [[12.07106781, 17.07106781, 12.07106781]],
                       np.array([10, 0., 0.]),
                       np.array([12.07106781, 17.07106781,
                                 0., 12.07106781, 0., 10.]),
                       np.array([10.11931299, 19.85794691, 21.57712079,
                                 19.48442822, 29.94069045,
                                 23.00413782]),
                       np.array([10.52224268,  13.74405405,  21.782617,  10.28175521,
                                 99.29457317,   5.30953205, 117.32861549,  23.76075308,
                                 2.54829623,  69.3944742])]
    sample_vecs_are_in = [True, True, False, False, False, False, False, False]
    sample_vecs_are_diff = [True, True, True, True, True, True, True, True]

if __name__ == '__main__':
    unittest.main()
    



import numpy as np
from cvxpy.reductions.solvers.solving_chain import construct_solving_chain
from .cones_utils import print_residuals, xsy2z, refine_py

""" Transform (x, y, s) returned from SCS to z-space. """
def xsy2z_support_certificates(x, s, y, b, c):
    # This function is called with the output (x, s, y) from SCS.
    # If the problem has a finite optimal value, SCS returns a triple (x, y, s)
    # satisfying the KKT-conditions for the conic LP.
    # If the problem is primal infeasible, then SCS returns (x, s) = nan and
    # a vector y such that b @ y = -1, A^T @ y = 0.
    # If the problem is dual infeasible, then SCS returns y = nan and (x, s) 
    # satisfying Ax + s = 0, c @ x = -1.

    # If the problem is primal infeasible.
    if np.any(np.isnan(x)):
        x = np.zeros_like(x)       
        s = np.zeros_like(s) 
        #y = -y/(b @ y)    # Not needed if b @ y = -1, but we want to make sure that
                          # our approximate certificate really satisfies b @ y = -1.
        z = xsy2z(x, s, y, tau=0., kappa=1.)
    # If the problem is dual infeasible.
    elif np.any(np.isnan(y)):
        y = np.zeros_like(y)
        #obj = c@x         # The following two lines are not needed if c @ x = -1,
        #x /= -obj         # but we want to make sure that our approximate
        #s /= -obj         # certificate really satisfies c @ x = -1.
        z = xsy2z(x, s, y, tau=0., kappa=1.) 
    # If the problem has a finite solution.
    else:
        z = xsy2z(x, s, y, tau=1., kappa=0.)

    return z

""" Parses cvxpy data object to standard conic LP form. """
def cvxpy_scs_to_coneref(data):
    A, b, c, dims = data['A'], data['b'], data['c'], data['dims']

    dims_dict = {}
    if int(dims.nonneg):
        dims_dict['l'] = int(dims.nonneg)
    if int(dims.zero):
        dims_dict['z'] = int(dims.zero)
    if int(dims.exp):
        dims_dict['ep'] = int(dims.exp)
    if len(dims.soc):
        dims_dict['q'] = list([int(el) for el in dims.soc])
    if len(dims.psd):
        dims_dict['s'] = list([int(el) for el in dims.psd])

    return A, b, c, dims_dict

""" This is the main function of the python interface. It has been designed to
    avoid the need for recompiling a problem that has already been compiled.
    * If cvxpy_problem has already been solved, its solution is refined.
    * If cvxpy_problem has been solved and refined, the refined solution is
          refined again.
    * If the cvxpy_problem has not been solved, the problem is solved using
      SCS and then the solution is refined. 

@param[in] cvxpy_problem    A cvxpy problem.
@param[in] ref_iter         Number of refinement steps.
@param[in] lsqr_iter        Number of LSQR iterations used for each refinement 
                            step.
@param[in] verbose_ref1      If true the refinement procedure writes out the
                             following quantities:
                          1. Primal and dual residual.
                          2. The complementary slackness measure
                             s^T y (here s denotes the primal slack variable 
                             in the standard conic LP form and y denotes
                             the dual variable).
                          3. The duality gap c^T x + b^T y.
                          4. The total time for SCS, and the total time for 
                             the refinement procedure.
@param[in] verbose_ref2      If true the refinement procedure writes out
                             the normalized residual N(z) (see eq. 8 of 
                             refinement paper) after each refinement iteration.
@param[in] warmstart, verbose and scs_opts are settings for SCS.
"""


def cvxpy_solve(cvxpy_problem, ref_iter=2, lsqr_iter=300, verbose_scs=True,
                scs_opts = {}, verbose_ref1 = True, verbose_ref2 = False,
                warm_start=False):
    
    # If the solving chain for the cvxpy problem has not been built, we build it 
    # and store it. We also solve the problem (which is on standard conic LP form).
    # Finally, we construct z corresponding to the (x, y, s)-solution obtained
    # from SCS.
    if len(cvxpy_problem._solver_cache) == 0:
        _candidates = {'qp_solvers': [], 'conic_solvers': ['SCS']}
        solving_chain =  construct_solving_chain(cvxpy_problem, candidates=_candidates)
        data, inverse_data = solving_chain.apply(cvxpy_problem)
        cvxpy_problem._solver_cache['data'] = data 
        cvxpy_problem._solver_cache['inverse_data'] = inverse_data
        cvxpy_problem._solver_cache['solving_chain'] = solving_chain
        scs_solution = solving_chain.solve_via_data(cvxpy_problem,
                            data=data, warm_start=warm_start, verbose=verbose_scs,
                            solver_opts=scs_opts)
        
        # If the problem is infeasible the primal variables x and s will be 
        # nan. If the problem is unbounded, the dual variable will be nan.
        x = scs_solution['x']
        y = scs_solution['y']
        s = scs_solution['s']
        total_time = scs_solution['info']['solve_time'] + \
                     scs_solution['info']['setup_time']
        cvxpy_problem._solver_cache['scs_total_time'] = total_time
        
        z = xsy2z_support_certificates(x, s, y, data['b'], data['c'])
        
        # Print residuals. There are residuals inside "scs_solution['info]"
        # but these residuals do not coincide with the true residuals. Maybe
        # they are computed for the scaled problem?
        A, b, c, dims_dict = cvxpy_scs_to_coneref(data)
        print("After SCS (total time = " + "{:.2e}".format(total_time/1000) + "s):")
        if np.any(np.isnan(x)):
            print_residuals(bTy=b@y, ATy_norm=np.linalg.norm(A.T @ y), 
                            primal_infeasible=True)
            cvxpy_problem._solver_cache['primal_infeasible'] = True
        elif np.any(np.isnan(y)):
            print_residuals(cTx=c@x, Axs_norm=np.linalg.norm(A @ x + s), 
                            dual_infeasible=True)
            cvxpy_problem._solver_cache['dual_infeasible'] = True
        else:
            # If SCS terminates because it has reached the maximum number of
            # iterations, cvxpy_problem._solver_cache does not contain the 'SCS' field
            # so we add it manually.
            if not ('SCS' in cvxpy_problem._solver_cache.keys()):
                cvxpy_problem._solver_cache['SCS'] = {'info': {'status_val': 2,
                'solve_time': scs_solution['info']['solve_time'], 
                'setup_time': scs_solution['info']['setup_time'],
                'iter': scs_solution['info']['iter']}}

            cvxpy_problem._solver_cache['SCS']['info']['res_pri_before_ref'] = np.linalg.norm(A @ x + s - b)
            cvxpy_problem._solver_cache['SCS']['info']['res_dual_before_ref'] = np.linalg.norm(A.T @ y + c)
            cvxpy_problem._solver_cache['SCS']['info']['gap_before_ref'] = c.T @ x + b.T @ y
            print_residuals(primal_residual = cvxpy_problem._solver_cache['SCS']['info']['res_pri_before_ref'],
                           dual_residual = cvxpy_problem._solver_cache['SCS']['info']['res_dual_before_ref'],
                           duality_gap = cvxpy_problem._solver_cache['SCS']['info']['gap_before_ref'],
                           finite_opt_val = True)
            cvxpy_problem._solver_cache['finite_opt_val'] = True

    # If the solving chain for the cvxpy problem has been built we reuse it.
    # We also reuse z from the previous round of refinement. 
    else:
        data = cvxpy_problem._solver_cache['data']
        inverse_data =  cvxpy_problem._solver_cache['inverse_data']
        solving_chain = cvxpy_problem._solver_cache['solving_chain']
        z = cvxpy_problem._solver_cache['z']

    # Parse data standard cone LP format.
    A, b, c, dims_dict = cvxpy_scs_to_coneref(data)

    refined_z, x, y, s, tau, kappa, info = \
                                  refine_py(A, b, c, dims_dict, z, ref_iter=ref_iter,
                                            lsqr_iter=lsqr_iter, verbose = verbose_ref2)
    cvxpy_problem._solver_cache['refine_time'] = info['ref_time']*1000 # In milliseconds
    
    if verbose_ref1:
        print("After refinement (ref time =" + "{:.2e}".format(info['ref_time']) + "s):")
        if 'primal_infeasible' in cvxpy_problem._solver_cache:
            print_residuals(bTy=b@y, ATy_norm=np.linalg.norm(A.T @ y), 
                            primal_infeasible=True)
        elif 'dual_infeasible' in cvxpy_problem._solver_cache:
             print_residuals(cTx=c@x, Axs_norm=np.linalg.norm(A @ x + s), 
                            dual_infeasible=True)
        elif 'finite_opt_val' in cvxpy_problem._solver_cache:
            print_residuals(primal_residual = info['primal_residual'],
                           dual_residual = info['dual_residual'],
                           duality_gap = info['duality_gap'],
                           finite_opt_val = True)


    # If the problem seems to have a finite solution. 
    if 'finite_opt_val' in cvxpy_problem._solver_cache: 
        # If SCS terminates because it has reached the maximum number of
        # iterations, cvxpy_problem._solver_cache does not contain the 'SCS' field
        # so we add it manually.
        #if not ('SCS' in cvxpy_problem._solver_cache.keys()):
        #    cvxpy_problem._solver_cache['SCS'] = {'info': {'status_val': 2,
        #    'solve_time': scs_solution['info']['solve_time'], 
        #    'setup_time': scs_solution['info']['setup_time'],
        #    'iter': scs_solution['info']['iter']}}

        # Update the information in SCS. 
        cvxpy_problem._solver_cache['SCS']['x'] = x            
        cvxpy_problem._solver_cache['SCS']['y'] = y 
        cvxpy_problem._solver_cache['SCS']['s'] = s
        cvxpy_problem._solver_cache['SCS']['info']['pobj'] = c @ x 
        cvxpy_problem._solver_cache['SCS']['info']['dobj'] = -b @ y      # Need the minus sign to compensate for earlier change of sign.
        cvxpy_problem._solver_cache['SCS']['info']['res_pri_after_ref'] = info['primal_residual']
        cvxpy_problem._solver_cache['SCS']['info']['res_dual_after_ref'] = info['dual_residual']
        cvxpy_problem._solver_cache['SCS']['info']['gap_after_ref'] = info['duality_gap']
        cvxpy_problem._solver_cache['SCS']['info']['comp_slack_after_ref'] = info['sTy']
        cvxpy_problem._solver_cache['z'] = refined_z
        cvxpy_problem.unpack_results(cvxpy_problem._solver_cache['SCS'], 
                                 solving_chain, inverse_data)
    # If primal infeasible problem.
    elif 'primal_infeasible' in cvxpy_problem._solver_cache:
        cvxpy_problem._solver_cache['primal_infeas_cert'] = {'y': y}
        cvxpy_problem._solver_cache['z'] = refined_z
    elif 'dual_infeasible' in cvxpy_problem._solver_cache:
        cvxpy_problem._solver_cache['dual_infeas_cert'] = {'x': x, 's': s}
    




import numpy as np
import sys   
import _coneref
from time import time
#import os  For finding process ID for debugging the C++ code.

ZERO = 'z'
POS = "l"
SOC = "q"
PSD = "s"
EXP = "ep"
EXP_DUAL = "ed"

# The ordering of CONES matches SCS.
CONES = [ZERO, POS, SOC, PSD, EXP, EXP_DUAL]

# Map from Python cones to C++ format.
CONE_MAP = {
    'z': _coneref.ConeType.ZERO,
    "l": _coneref.ConeType.POS,
    "q": _coneref.ConeType.SOC,
    "s": _coneref.ConeType.PSD,
    "ep": _coneref.ConeType.EXP,
    "ed": _coneref.ConeType.EXP_DUAL
}

# The following two functions are used to convert a dictionary of cones
# to C++ format.
def parse_cone_dict_cpp(cone_list):
    return [_coneref.Cone(CONE_MAP[cone], [l] if not isinstance(l, (list, tuple)) else l)
            for cone, l in cone_list]
def parse_cone_dict(cone_dict):
    return [(cone, cone_dict[cone]) for cone in CONES if cone in cone_dict]

# Evaluates N(z) defined in eq. 8 of the refinement paper.
def evaluate_normalized_res_map(z, A, b, c, cones):
    _cones = parse_cone_dict(cones)
    _cones_parsed = parse_cone_dict_cpp(_cones)
    m, n = A.shape
    # These caches are not really needed.
    q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
         = make_prod_cone_cache(cones)
    normalized_res = _coneref.residual_map_python_friendly(A, b, c, _cones_parsed,
                      z, q_cache, eval_cache, evec_cache, ep_cache, ed_cache, n, m)

    return normalized_res

# Compute x, s, y, kappa, tau from z."""
#def _z2xsykappatau(z, cones, n, m):
#    _cones = parse_cone_dict(cones)
#    _cones_parsed = parse_cone_dict_cpp(_cones)

     # These caches are not actually needed.
#    q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
#        = make_prod_cone_cache(cones)
#    u = _coneref.projection_embedded_cone(z, _cones_parsed, q_cache, eval_cache, 
#                                        evec_cache, ep_cache, ed_cache, n, m)
#    v = u - z
#    x, s, y, tau, kappa = uv2xsytaukappa(u, v, n)
#    return x, s, y, tau, kappa

# The following functions are used to set up caches for all cone types.
def make_prod_cone_cache(dim_dict):
    return _make_prod_cone_cache(np.array(dim_dict['q'] if 'q'
                                          in dim_dict else [], dtype=np.int64),
                                 np.array(dim_dict['s'] if 's' in
                                          dim_dict else [], dtype=np.int64),
                                 np.array(dim_dict['ep'] if 'ep' in
                                          dim_dict else 0, dtype=np.int64),
                                 np.array(dim_dict['ed'] if 'ed' in
                                          dim_dict else 0, dtype=np.int64))

def _make_prod_cone_cache(second_ord, semi_def, ep, ed):
    q_cache = np.zeros(np.sum(second_ord))

    eval_cache = np.zeros(np.sum(semi_def))
    evec_cache = np.zeros(np.sum(semi_def**2))

    ep_cache = np.zeros((3*ep, ))
    ed_cache = np.zeros((3*ed, ))

    return q_cache, eval_cache, evec_cache, ep_cache, ed_cache

# Given x, s, y this function returns u = [x y tau], v = [0, s, kappa].
def xsy2uv(x, s, y, tau=1., kappa=0.):
    n = len(x)
    m = len(s)
    u = np.empty(m + n + 1)
    v = np.empty_like(u)
    u[:n] = x
    u[n:-1] = y
    u[-1] = tau
    v[:n] = 0
    v[n:-1] = s
    v[-1] = kappa
    return u, v

# Returns z = u - v"""
def xsy2z(x, s, y, tau=1., kappa=0.):
    u, v = xsy2uv(x, s, y, tau, kappa)
    return u - v


# Given u and v this function returns x, s, y, tau, kappa.
def uv2xsytaukappa(u, v, n):
    tau = np.float(u[-1])
    kappa = np.float(v[-1])
    x = u[:n] / tau if tau > 0 else u[:n] / kappa
    y = u[n:-1] / tau if tau > 0 else u[n:-1] / kappa
    s = v[n:-1] / tau if tau > 0 else v[n:-1] / kappa
    return x, s, y, tau, kappa

#  Refines an approximate solution to the conic linear program
#   min. c^T x subject to Ax + s = b, s \in \mathcal{K}.
def refine_py(A, b, c, cones, z, ref_iter = 2, lsqr_iter = 30, 
              verbose = False):
    _cones = parse_cone_dict(cones)
    _cones_parsed = parse_cone_dict_cpp(_cones)
    m, n = A.shape 
    q_cache, eval_cache, evec_cache, ep_cache, ed_cache \
        = make_prod_cone_cache(cones) # These caches are not actually needed.
    info = {}
    
    # Refine the solution
    tic = time()
    refined_z = _coneref.refine(A, b, c, _cones_parsed, z, n, m, ref_iter, lsqr_iter,
                             verbose)

    # Recover x, y, s from z. Requires an additional projection.
    u = _coneref.embedded_cone_Pi(refined_z, _cones_parsed, q_cache, eval_cache,
                                evec_cache, ep_cache, ed_cache, n, m)
    v = u - refined_z    
    x, s, y, tau, kappa = uv2xsytaukappa(u, v, n)
    
    
    # Store some quantities
    info['ref_time'] = time() - tic
    info['primal_residual'] = np.linalg.norm(A @ x + s - b)
    info['dual_residual'] = np.linalg.norm(A.T @ y + c)
    info['sTy'] = s @ y 
    info['duality_gap'] = c.T @ x + b.T @ y

    return refined_z, x, y, s, tau, kappa, info

def print_residuals(primal_residual = -1, dual_residual = -1, duality_gap = -1,
                    bTy = -1, ATy_norm = -1, Axs_norm = -1, cTx = -1,
                    primal_infeasible = False, dual_infeasible = False,
                    finite_opt_val = False):

    if finite_opt_val:                
        print("Primal residual/dual residual/duality_gap:",
                    "{:.4e}".format(primal_residual), " ",
                    "{:.4e}".format(dual_residual), " ",
                    "{:.4e}".format(duality_gap))
    elif primal_infeasible:
        print("b @ y:", bTy)
        print("||A^T y||:", ATy_norm)
    elif dual_infeasible:
        print("c @ x:", "{:.4e}".format(cTx))
        print("||Ax + s||:", "{:.4e}".format(Axs_norm))


import distutils.sysconfig
import distutils.version
from glob import glob
import os
import platform
from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext
import subprocess
import sys

with open("README.md", "r") as fh:
    long_description = fh.read()

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __init__(self, user=False):
        try:
            import pybind11
        except ImportError:
            if subprocess.call([sys.executable, '-m', 'pip', 'install', 'pybind11']):
                raise RuntimeError('pybind11 install failed.')
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)



def march_native():
    try:
        bit = os.environ["MARCH_NATIVE"]
        if bit == '1':
            return ["-march=native"]
        else:
            return []
    except KeyError:
        return []


_coneref = Extension(
    '_coneref',
    glob("cpp/src/*.cpp"),
    include_dirs=[
        get_pybind_include(),
        get_pybind_include(user=True),
        "cpp/external/eigen",
        "cpp/external/eigen/Eigen",
        "cpp/include",
    ],
    language='c++',
    extra_compile_args=["-O3", "-std=c++11"] + march_native()
)


def is_platform_mac():
    return sys.platform == 'darwin'


# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distutils behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py. This behavior is
# motivated by Apple dropping support for libstdc++.
if is_platform_mac():
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        current_system = distutils.version.LooseVersion(platform.mac_ver()[0])
        python_target = distutils.version.LooseVersion(
            distutils.sysconfig.get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < '10.9' and current_system >= '10.9':
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'

ext_modules = [_coneref]

setup(
    name='coneref',
    version="0.2.2",
    author="Daniel Cederberg, Stephen Boyd",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    setup_requires=['pybind11 >= 2.4'],
    install_requires=[
        "numpy >= 1.15",
        "scs >= 2.0.2",  # 2.0.2 is the oldest version on conda forge
        "scipy >= 1.1.0",
        "pybind11 >= 2.4",
        "cvxpy >= 1.1.0"],
    url="https://github.com/dance858/coneref",
    ext_modules=ext_modules,
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
)
